!Calculating Ne, L, variance of family size, probabilities of sibship & 
!  parentage for dioecious diploid populations with overlapping generations
!
!Call external program POAssignment for parentage assignment analysis
!Read in a text data file named "AgeStructure.dat"
!Output results in 3 files: OutputFileName.Paternity, 
!  OutputFileName.Maternity, OutputFileName.Ne, where OutputFileName 
!  is defined in the 2nd line of data file "AgeSexGtype.dat"
!
!
!Compile options
!  Ifort: Ifort AgeStructure.f90 /O2 /exe:AgeStructure; Ifort POAssignment.f90 /O2 /exe:POAssignment
!  g95: g95 AgeStructure.f90 -O3 -o AgeStructure; g95 POAssignment.f90 -O3 -o POAssignment
!    Use 'SET G95_MEM_SEGMENTS=0' &
!    before running executable to disable warning on memery
!
Module AgeVariables
  IMPLICIT NONE
  INTEGER, PARAMETER :: DP = KIND(1.0D0)              ! Double Precision
  INTEGER, PARAMETER :: I8R = SELECTED_REAL_KIND(9, 1000)      
  INTEGER, PARAMETER :: MaxIntNumber = 2147483647
  Character(Len = 8), Parameter :: Build = '20160719'  !Year-Month-Day
  Character(Len = 7), Parameter :: Version = '1.1'
  REAL(DP), PARAMETER :: TinyValue = -1.E300_DP, Small = 1.E-300_DP
  Integer :: NumAgeClass(0 : 2), MaxNumAgeClass, idum, NumLoci, &
    SmplSize, NumCI95Sample, MaxObsNumAllele, FamilySizeDist, MinParentAge(2)
  Integer, Allocatable :: AgeClassSize(:, :), IndivSex(:), IndivAge(:), &
    ObsNumAllele(:), Dominant(:), SexAgeSize(:, :), SexAgeSize0(:, :) 
  Integer :: NeEstimate(0 : 2), SamplingPrbKnown, CrrtOS
  REAL(DP), Allocatable :: CP(:, :), AlleleFre(:, :), DropRate(:), OtherRate(:)
  REAL(DP) :: GIEstimate(0 : 2, 0 : 2), Var(3, 2, 0 : 2), PropPrntSampled(2)
  REAL(DP), Allocatable :: Estimate(:, :, :), ReproductionRate(:), &
    SamplingPrb(:, :), ObsAlleleFre(:, :), LogFactorial(:), &
    Factor(:), GammlnSave(:, :), gammlnSmpl(:), SumNumOff(:), &
    SampleFre(:, :)
  Character(Len = 20), Allocatable :: OffspringID(:), CndParentID(:, :), &
    MarkerID(:), SortCharacter(:)    
  Character(Len = 100) :: FName(2), DataSetName   
  TYPE Member
    Integer, Pointer :: M(:, :)
  END TYPE Member  
  Type(Member), Allocatable :: AgeClass(:, :)  
  TYPE GenotypeData
    Integer, Pointer :: G(:, :, :)
  END TYPE GenotypeData   
  Type(GenotypeData), Allocatable :: GType(:, :) 
  TYPE MemberVector
    Integer, Pointer :: I(:)
  END TYPE MemberVector
  Type(MemberVector), Allocatable :: Death(:, :)       
  INTEGER, Allocatable :: NumOff(:, :), NumOffSum(:), &
    NumOffSum2(:), SampledCounts(:), PODyadCounts(:, :), &
    PODyadCounts0(:, :, :), NumOff0(:, :, :)
  INTEGER :: NumOff0Bound(2)
  TYPE RealVector
    REAL(DP), Pointer :: R(:)
  END TYPE RealVector
  TYPE(RealVector), Allocatable :: SamplingPrbPrior(:)
  REAL(DP), Pointer :: SamplingPrbPrt(:)
  Character(Len = 50) :: OutputFileName
  CHARACTER(Len = 8) :: Date2(2)
  CHARACTER(Len = 10) :: Time2(2)  
  CHARACTER(Len = 2) :: ConfidenceLevel
  REAL(DP) :: Ran2  
  EXTERNAL :: Ran2
END MODULE AgeVariables
!
Module PowellVariables
  USE AgeVariables, ONLY : DP
  IMPLICIT NONE
  Integer :: NCOM, NDIM1, NDIM2
  REAL(DP), Allocatable :: pcom(:), xicom(:)
End Module PowellVariables
!
Module FuncModule
  Use AgeVariables   
  USE PowellVariables
  IMPLICIT NONE
  CONTAINS
FUNCTION FUNC(X)
  REAL(DP) :: func, Z(NDIM2)
  REAL(DP), Target :: X(NCOM)
  REAL(DP) :: F, Y, A, Gammln
  INTEGER :: I, J, K, L, M 
  EXTERNAL :: Gammln
  K = NDIM1 + NDIM2  
  IF(SamplingPrbKnown /= 1) SamplingPrbPrt => X(K + 1 : NCOM)
  Z(1 : NDIM1) = X(1 : NDIM1)
  Z(NDIM2) = 1._DP - Sum(X(1 : NDIM1))
  IF(ANY(Z(:) <= 0.0_DP) .OR. ANY(X(NDIM2 : K) < 1.0_DP) .OR. &
     ANY(X(NDIM1 + 2 : K) > X(NDIM2 : K - 1) + 1.E-5_DP) .OR. &
     ANY(X(NDIM2 : K) > 1.E5_DP) .OR. &
     ANY(SamplingPrbPrt( : ) > 1.0_DP) .OR. &
     ANY(SamplingPrbPrt( : ) < 1.E-6_DP)) THEN
    FUNC = -TinyValue
    RETURN
  END IF   
  F = .0_DP
!Multivariate hypergeometric distribution 
  A = 0._DP
  DO I = 1, NDIM2
    Y = X(NDIM1 + I) * SamplingPrbPrt(I)
    If(Y <= SampledCounts(I)) THEN
      FUNC = -TinyValue
      RETURN
    END IF 
    IF(ABS(GammlnSave(1, I) - Y) > 1.E-30_DP) THEN
      GammlnSave(1, I) = Y
      GammlnSave(2, I) = gammln(1._DP + Y) - gammlnSmpl(I) - &
        gammln(1._DP + Y - SampledCounts(I))
    END IF
    F = F + GammlnSave(2, I)
    A = A + Y
  END DO 
  IF(ABS(GammlnSave(1, 0) - A) > 1.E-30_DP) THEN
    GammlnSave(1, 0) = A
    GammlnSave(2, 0) = -gammln(1._DP + A) + gammlnSmpl(0) + &
      gammln(1._DP + A - SampledCounts(0))
  END IF
  F = F + GammlnSave(2, 0)
!C Distribution of family size per parent: Poisson distribution
  L = UBound(NumOff, 1)
  IF(L > 0) THEN
    DO I = NDIM2, K 
      M = I - NDIM1
      IF(NumOffSum(M) < 1) CYCLE 
      Y = NumOffSum(M) / (SamplingPrbPrt(I - NDIM1) * X(I))    
      F = F + Log(Y) * NumOffSum(M) - Y * NumOffSum2(M) - SumNumOff(M)
    END DO
  END IF
!C Prb an individual of age j is assigned parentage to an offspring of age i(i<j)
  DO I = 1, NDIM2  
    Z(I) = Z(I) / X(NDIM1 + I)
  END DO   
  DO I = 0, NDIM1
    A = 0._DP
    M = PODyadCounts(I, 0)
    DO J = I + 1, NDIM2
      IF(PODyadCounts(I, J) > 0) THEN
        Y = Z(J - I) * SamplingPrbPrt(J) * X(NDIM1 + J)   
        F = F + Log(Y) * PODyadCounts(I, J) - LogFactorial(PODyadCounts(I, J))
        A = A + Y
        M = M - PODyadCounts(I, J)
      ELSE
        A = A + Z(J - I) * SamplingPrbPrt(J) * X(NDIM1 + J) 
      END IF         
    END DO
    IF(A >= 1._DP) THEN
      FUNC = -TinyValue
      RETURN
    END IF 
    F = F + Log(1._DP - A) * M - &
      LogFactorial(M) + LogFactorial(PODyadCounts(I, 0))
  END DO
  FUNC = -F
END FUNCTION FUNC
END Module FuncModule
!
Program MAIN
  USE AgeVariables
  IMPLICIT NONE
  EXTERNAL :: ReadData, initiate, EstimateByMarker, &
    Output, RemoveFile, CopyFile, ShowVersion
  CALL ShowVersion
  CALL ReadData
  CALL RemoveFile('AgeSexGtype.dat')
  CALL CopyFile('AgeStructure.dat', 'AgeSexGtype.dat')
  CALL initiate
  CALL EstimateByMarker(0, NeEstimate(0), GIEstimate(0 : 2, 0), &
    Var(:, :, 0), NumAgeClass(0), Estimate(:, :, 0))
  IF(NumCI95Sample > 0) CALL FindCI95
  CALL Output
END Program MAIN
!
Subroutine ShowVersion
  USE AgeVariables
  IMPLICIT NONE
  INTEGER :: I
  Character(Len = 8) :: Line * 82
  FORALL(I = 1 : 82) Line(I : I) = '-'
  WRITE(*, '(/A/A)') Line, 'AgeStructure, Version ' // &
    TRim(AdjustL(Version)) // ', Build ' // Build
  WRITE(*, '(A)') 'Copyright (C) by Jinliang Wang, Institute of Zoology, ' // &
    'Zoological Society of London'
  WRITE(*, '(A/A//)') 'Email: jinliang.wang@ioz.ac.uk', Line
  CALL Sleeping(1.0)
END Subroutine ShowVersion
!
Subroutine Sleeping(snd)
  implicit none
  real :: snd, dt
  integer :: t(8), ms1, ms2
  dt = snd * 1000.0     ! desired sleep interval in [ms]
  call date_and_time(values=t)
  ms1=(t(5)*3600+t(6)*60+t(7))*1000+t(8)
  do
    call date_and_time(values=t)
    ms2=(t(5)*3600+t(6)*60+t(7))*1000+t(8)
    if(ms2-ms1>=dt)exit
  enddo
End Subroutine Sleeping
!
Subroutine RemoveFile(FileName)
  USE AgeVariables
  CHARACTER(*) :: FileName
  Logical :: ExistQ
  Inquire(FILE = Trim(AdjustL(FileName)), Exist = ExistQ)
  IF(.NOT. ExistQ) RETURN
  OPEN(10, File = Trim(AdjustL(FileName)))
  CLOSE(10, STATUS = 'DELETE')
END Subroutine RemoveFile
!
Subroutine CopyFile(Fname1, Fname2)
  USE AgeVariables
  CHARACTER(*) :: Fname1, Fname2
  CHARACTER(Len = 2000) :: C2000
  Logical :: ExistQ
  Inquire(FILE = Trim(AdjustL(Fname1)), Exist = ExistQ)
  IF(.NOT. ExistQ) RETURN
  OPEN(10, File = Trim(AdjustL(Fname1)))
  OPEN(11, File = Trim(AdjustL(Fname2)))
  DO
    READ(10, '(A2000)', ERR=9,End=9) C2000
    WRITE(11, '(A)') Trim(C2000)
  END DO
9 CLOSE(10)
  CLOSE(11)
END Subroutine CopyFile
!
Subroutine ReadData
  USE AgeVariables
  IMPLICIT NONE
  INTEGER :: I, J, K, L, M, IOS, NumLines
  INTEGER, Allocatable :: PopAllele(:, :), IndivGType(:, :, :), &
    Genotype(:), Idx(:)
  REAL(DP) :: X
  CHARACTER(Len = 999) :: ErrorMessage
  EXTERNAL :: StopOnError
  CALL DATE_AND_TIME(Date2(1), Time2(1))
  ErrorMessage = 'Error in reading file AgeStructure.dat in Subroutine ReadData'
  OPEN(UNIT=10,FILE='AgeStructure.dat',IOSTAT=IOS,ERR=9,STATUS='OLD')
  ErrorMessage = 'Error in reading line 1, Dataset Name'   
  READ(10,*,IOSTAT=IOS,ERR=9,End=9)DataSetName
  ErrorMessage = 'Error in reading line 2, Output File Name'  
  READ(10,*,IOSTAT=IOS,ERR=9,End=9)FNAME(1)
  ErrorMessage = 'Error in reading line 3, Offspring Sample Size'  
  READ(10,*,IOSTAT=IOS,ERR=9,End=9)SmplSize
  IF(SmplSize < 1) THEN
    ErrorMessage = &
      'The sample size of offspring should be greater than 0!'
    GOTO 9
  END IF
  ErrorMessage = 'Error in reading line 4, Maximal age for males & females'      
  READ(10, *,IOSTAT=IOS,ERR=9,End=9)NumAgeClass(1 : 2) 
  ErrorMessage = '' 
  IF(NumAgeClass(1) < 1) THEN
    ErrorMessage = 'The maximal age for males is too small!'
  ELSE IF(NumAgeClass(1) > 50) THEN
    ErrorMessage = 'The maximal age for males is too large!'
  ELSE IF(NumAgeClass(2) < 1) THEN
    ErrorMessage = 'The maximal age for females is too small!'
  ELSE IF(NumAgeClass(2) > 50) THEN
    ErrorMessage = 'The maximal age for females is too large!'
  END IF 
  IF(ErrorMessage /= '') GOTO 9
  NumAgeClass(0) = NumAgeClass(1) + NumAgeClass(2)
  ErrorMessage = 'Error in reading line 5, ' // &
    'Minimal reproductive age for males & females'      
  READ(10, *,IOSTAT=IOS,ERR=9,End=9) MinParentAge(1 : 2)
  ErrorMessage = '' 
  IF(MinParentAge(1) < 1) THEN
    ErrorMessage = 'The minimal reproductive age for males is too small!'
  ELSE IF(MinParentAge(1) > NumAgeClass(1)) THEN
    ErrorMessage = 'The minimal reproductive age for males is too large!'
  ELSE IF(MinParentAge(2) < 1) THEN
    ErrorMessage = 'The minimal reproductive age for females is too small!'
  ELSE IF(MinParentAge(2) > NumAgeClass(2)) THEN
    ErrorMessage = 'The minimal reproductive age for females is too large!'
  END IF 
  IF(ErrorMessage /= '') GOTO 9
  ErrorMessage = 'Error in reading line 6, assumed family size distribution!' 
  READ(10, *,IOSTAT=IOS,ERR=9,End=9)FamilySizeDist
  IF(FamilySizeDist /= 0 .AND. FamilySizeDist /= 1) THEN
    ErrorMessage = 'The indicator for assumed family size distribution' &
      // ' must be 1/0 for Poisson/Other distribution!'
    GOTO 9
  END IF
  ErrorMessage = 'Error in reading line 7, sampling proportion' // &
    ' known/unknown for each age class!' 
  READ(10, *,IOSTAT=IOS,ERR=9,End=9) SamplingPrbKnown
  IF(SamplingPrbKnown /= 0 .AND. SamplingPrbKnown /= 1) THEN
    ErrorMessage = 'The indicator for sampling proportion' &
      // ' must be 1/0 for known/unknown!'
    GOTO 9
  END IF
  IF(SamplingPrbKnown == 1) THEN
    ErrorMessage = 'Error in reading known sampling proportions of each age class'
    Allocate(SamplingPrbPrior(2))
    DO K = 1, 2
      Allocate(SamplingPrbPrior(K)%R(NumAgeClass(K)))
      READ(10, *,IOSTAT=IOS,ERR=9,End=9)(SamplingPrbPrior(K)%R(J), J = 1, NumAgeClass(K))
      DO J = 1, NumAgeClass(K)
        IF(SamplingPrbPrior(K)%R(J) > 1._DP .OR. &
           SamplingPrbPrior(K)%R(J) < 0._DP) THEN
          ErrorMessage = 'Invalid values for known sampling proportions!'
          GOTO 9
        END IF
      END DO
    END DO
    NumLines = 9
  ELSE
    NumLines = 7
  END IF
  NumLines = NumLines + 1
  WRITE(ErrorMessage, '(A,I4, A)') 'Error in reading line', NumLines, &
    ', number of bootstrapping samples!'
  READ(10, *,IOSTAT=IOS,ERR=9,End=9)NumCI95Sample
  ErrorMessage = ''
  IF(NumCI95Sample < 0) THEN
    ErrorMessage = 'The number of bootstrapping samples must be >=0!'
  ELSE IF(NumCI95Sample > 100000) THEN
    ErrorMessage = 'The number of bootstrapping samples is too big!'
  END IF
  IF(ErrorMessage /= '') GOTO 9
  NumLines = NumLines + 1
  WRITE(ErrorMessage, '(A,I4, A)') 'Error in reading line', &
    NumLines, ', Number of Loci'      
  READ(10,*,IOSTAT=IOS,ERR=9,End=9)NumLoci
  IF(NumLoci < 1) THEN
    ErrorMessage = 'The number of loci should be greater than 0!'
    GOTO 9
  END IF
  NumLines = NumLines + 1
  WRITE(ErrorMessage, '(A,I4, A)') 'Error in reading line', &
    NumLines, ', Confidence level for parentage assignment'      
  READ(10,*,IOSTAT=IOS,ERR=9,End=9)I
  IF(I < 80 .AND. I > 95 .AND. Mod(I, 5) /= 0) THEN
    ErrorMessage = 'The confidence level for parentage' // &
      ' assignment must be one of {80, 85, 90, 95}!'
    GOTO 9
  END IF
  Write(ConfidenceLevel, '(I2)') I 
  Allocate(Dominant(NumLoci), ObsNumAllele(NumLoci))
  NumLines = NumLines + 1
  WRITE(ErrorMessage, '(A,I4, A)') 'Error in reading line', &
    NumLines, ', indicator population allele frequency'  
  NumLines = NumLines + 1
  Allocate(MarkerID(NumLoci))
  WRITE(ErrorMessage, '(A,I4, A)') 'Error in reading line', &
    Numlines, ', Marker IDs'   
  READ(10,*,IOSTAT=IOS,Err=9,End=9) (MarkerID(I), I = 1, NumLoci)
  DO I = 1, NumLoci
    MarkerID(I) = Trim(AdjustR(MarkerID(I)))
  END DO 
  IF(Allocated(SortCharacter)) Deallocate(SortCharacter)   
  ALLOCATE(Idx(NumLoci), SortCharacter(NumLoci))
  DO I = 1, NumLoci 
    SortCharacter(I) = MarkerID(I)
  END DO
  CALL indexCh(NumLoci, Len(SortCharacter(1)), idx)  
  DO I = 2, NumLoci
    IF(MarkerID(Idx(I)) == MarkerID(Idx(I - 1))) THEN
      WRITE(ErrorMessage,'(A,I8,A,A,A,I8)') 'Marker', Idx(I), &
        ' has the same ID, ', Trim(MarkerID(Idx(I))), ', as Marker', Idx(I - 1)
      GOTO 9
    END IF
  END DO
  NumLines = NumLines + 1
  WRITE(ErrorMessage, '(A,I4, A)') 'Error in reading line', &
    Numlines, ', indicators for marker types per locus'      
  READ(10,*,IOSTAT=IOS,Err=9,End=9) Dominant(1 : NumLoci)
  IF(ANY(Dominant(:) > 1) .OR. ANY(Dominant(:) < 0)) THEN
    ErrorMessage = 'Marker type should be either 1 (dominant) or 0 (codominant)!'
    GOTO 9
  END IF  
  Allocate(DropRate(NumLoci), OtherRate(NumLoci))
  NumLines = NumLines + 1
  WRITE(ErrorMessage, '(A,I4, A)') 'Error in reading line', &
    Numlines, ', Dropout Rate per locus'    
  READ(10,*,IOSTAT=IOS,Err=9,End=9) DropRate(1 : NumLoci) !Drop rate for each locus
  NumLines = NumLines + 1
  WRITE(ErrorMessage, '(A,I4, A)') 'Error in reading line', &
    Numlines, ', Other Error Rate per locus'        
  READ(10,*,IOSTAT=IOS,Err=9,End=9) OtherRate(1 : NumLoci) !OtherErrorRate for each locus
  IF(ANY(DropRate(:) <= -Small) .OR. ANY(OtherRate(:) <= -Small) .OR. &
    ANY(DropRate(:) > 1._DP) .OR. ANY(OtherRate(:) > 1._DP)) THEN
    ErrorMessage = 'Marker error rate is out of the range [0, 1]!'
    GOTO 9
  END IF
  ALLOCATE(IndivGType(2, NumLoci, SmplSize), Genotype(NumLoci), &
    OffspringID(SmplSize), IndivSex(SmplSize), IndivAge(SmplSize), STAT = I)
  IF(I /= 0) Call StopOnError('Error in allocating arrays in subroutine ReadData')
  DO I = 1, NumLoci
    Genotype(I) = 2 - Dominant(I)
  END DO    
  DO I = 1, SmplSize
    NumLines = NumLines + 1
    WRITE(ErrorMessage, '(A,I4, A)') 'Error in reading line', &
      Numlines, ', Data or data format error in offspring ID/Sex/Age/genotypes'   
    READ(10,*,IOSTAT=IOS,Err=9,End=9) OffspringID(I), IndivSex(I), IndivAge(I), &
      (IndivGType(1 : Genotype(J), J, I), J = 1, NumLoci)
  END DO
  IF(Allocated(PopAllele))Deallocate(PopAllele)
  Allocate(PopAllele(SmplSize * 2, 1))
  MaxObsNumAllele = 0
  DO J = 1, NumLoci
    M = 0
    DO I = 1, SmplSize
      IF(ANY(IndivGType(1 : 2, J, I) < 1)) CYCLE
      DO K = 1, 2
        IF(ANY(IndivGType(K, J, I) == PopAllele(1 : M, 1))) CYCLE
        M = M + 1
        PopAllele(M, 1) = IndivGType(K, J, I)
      END DO
    END DO
    ObsNumAllele(J) = M
    IF(M > MaxObsNumAllele) MaxObsNumAllele = M
  END DO
  IF(Allocated(ObsAlleleFre)) Deallocate(ObsAlleleFre)
  ALLOCATE(ObsAlleleFre(NumLoci, MaxObsNumAllele))
  DO J = 1, NumLoci
    ObsAlleleFre(J, :) = 0._DP
    M = 0
    DO I = 1, SmplSize
      IF(ANY(IndivGType(1 : 2, J, I) < 1)) CYCLE
      DO K = 1, 2
        IOS = 0
        DO L = 1, M
          IF(IndivGType(K, J, I) == PopAllele(L, 1)) THEN
            ObsAlleleFre(J, L) = ObsAlleleFre(J, L) + 1._DP
            IOS = 1
            EXIT
          END IF
        END DO
        IF(IOS == 0) THEN
          M = M + 1
          PopAllele(M, 1) = IndivGType(K, J, I)
          ObsAlleleFre(J, L) = 1._DP
        END IF
      END DO
    END DO
    X = 1._DP / Sum(ObsAlleleFre(J, :))
    DO I = 1, ObsNumAllele(J)
      ObsAlleleFre(J, I) = ObsAlleleFre(J, I) * X
    END DO
  END DO 
  READ(10,*,IOSTAT=IOS,Err=9,End=9) PropPrntSampled(1 : 2)
  CLOSE(10)
  RETURN
9 CALL StopOnError(ErrorMessage)
END Subroutine ReadData
!
Subroutine Initiate
  USE AgeVariables
  IMPLICIT NONE
  CHARACTER(LEN = 999) :: pathstring
  IDUM = -1234
  Allocate(Estimate(NumAgeClass(0), 4, 0 : 2))
  CALL get_environment_variable("PATH", pathstring)
  IF(pathstring(1:1) == '/') THEN
    CrrtOS = 2     !Linux/Apple
  ELSE
    CrrtOS = 1     !Windows
  END IF
END Subroutine initiate
!
Subroutine FindCI95
  USE AgeVariables
  IMPLICIT NONE
  INTEGER :: NumAge(0 : 2)
  REAL(DP) :: P(NumAgeClass(0)), RM, X
  REAL(DP), Allocatable :: GIBoots(:, :), VBoots(:, :, :), EstBoots(:, :, :)
  INTEGER, Allocatable :: NeBoots(:), Idx(:)
  INTEGER :: I, J, K, L, M, JR, MaxNumAge
  EXTERNAL :: Indexx
  NumAge(0 : 2) = NumAgeClass(0 : 2)
  P = Estimate(:, 1, 0)
  MaxNumAge = Maxval(NumAge(1 : 2))
  Allocate(ReproductionRate(NumAge(0)), AgeClassSize(0 : MaxNumAge, 2), &
    SamplingPrb(0 : MaxNumAge, 2))
!C Set parameter values to the estimates
  DO I = 1, 2
    IF(I == 1) THEN
      AgeClassSize(1 : NumAge(I), I) = NINT(Estimate(1 : NumAge(1), 2, 0))
      SamplingPrb(1 : NumAge(I), I) = Estimate(1 : NumAge(I), 3, 0)
    ELSE
      AgeClassSize(1 : NumAge(I), I) = &
        NINT(Estimate(1 + NumAge(1) : NumAge(1) + NumAge(2), 2, 0))
      SamplingPrb(1 : NumAge(I), I) = &
        Estimate(1 + NumAge(1) : NumAge(0), 3, 0)
    END IF
    AgeClassSize(0, I) = Max(AgeClassSize(1, I), SexAgeSize(0, I))
    SamplingPrb(0, I) = SamplingPrb(1, I)
  END DO
  DO I = 1, NumAge(0)
    ReproductionRate(I) = P(I)
  END DO
!C Do Boots trapping
  IF(Allocated(NeBoots)) Deallocate(NeBoots, GIBoots, VBoots, EstBoots)
  Allocate(NeBoots(NumCI95Sample), GIBoots(0 : 2, NumCI95Sample), &
    VBoots(3, 2, NumCI95Sample), EstBoots(NumAgeClass(0), 4, NumCI95Sample)) 
  Allocate(SampleFre(0 : MaxNumAgeClass, 2))
  RM = 0._DP
  DO I = 1, 2
    DO J = 0, NumAgeClass(I)
      SampleFre(J, I) = SexAgeSize(J, I)
      RM = RM + SampleFre(J, I)
    END DO
  END DO
  RM = 1._DP / RM
  X = -1._DP
  DO I = 1, 2
    DO J = 0, NumAgeClass(I)
      IF(X < 0._DP) THEN
        SampleFre(J, I) = SampleFre(J, I) * RM
      ELSE
        SampleFre(J, I) = X + SampleFre(J, I) * RM
      END IF
      X = SampleFre(J, I)
    END DO
  END DO
  SampleFre(NumAgeClass(2), 2) = 2._DP   
  DO JR = 1, NumCI95Sample
    IF(Mod(JR, 5) == 0) write(*,*)'Bootstraping Replicate=', JR    
    CALL Simulate
    CALL WriteColonyData  
    CALL EstimateByMarker(JR, NeBoots(JR), GIBoots(:, JR), &
      VBoots(:, :, JR), NumAgeClass(0), EstBoots(:, :, JR))
  END DO
  IF(Allocated(Idx)) Deallocate(Idx)
  Allocate(Idx(NumCI95Sample))
  I = Max(1, NINT(0.025_DP * NumCI95Sample))
  J = NINT(0.975_DP * NumCI95Sample) 
  M = Max(NumCI95Sample / 2, 1)   
  CALL Indexx(NumCI95Sample, Dble(NeBoots(:)), Idx)  
  NeEstimate(1) = NeBoots(Idx(I))
  NeEstimate(2) = NeBoots(Idx(J))
  IF(NeEstimate(0) <= NeEstimate(1) .OR. NeEstimate(0) >= NeEstimate(2)) THEN
    X = Sum(1.D0 / NeBoots(1 : NumCI95Sample))
    NeEstimate(0) = NINT(NumCI95Sample / X)
  END IF
  DO K = 0, 2
    CALL Indexx(NumCI95Sample, GIBoots(K, :), Idx)
    GIEstimate(K, 1) = GIBoots(K, Idx(I))
    GIEstimate(K, 2) = GIBoots(K, Idx(J))
    IF(GIEstimate(K, 0) <= GIEstimate(K, 1) .OR. &
       GIEstimate(K, 0) >= GIEstimate(K, 2)) &
      GIEstimate(K, 0) = GIBoots(K, Idx(M))
  END DO  
  DO K = 1, 2
    DO L = 1, 3
      CALL Indexx(NumCI95Sample, VBoots(L, K, :), Idx)
      Var(L, K, 1) = VBoots(L, K, Idx(I))
      Var(L, K, 2) = VBoots(L, K, Idx(J))
      IF(Var(L, K, 0) <= Var(L, K, 1) .OR. &
         Var(L, K, 0) >= Var(L, K, 2)) Var(L, K, 0) = VBoots(L, K, Idx(M))
    END DO
  END DO
  DO L = 1, NumAgeClass(0)
    DO K = 1, 4
      CALL Indexx(NumCI95Sample, EstBoots(L, K, :), Idx)
      Estimate(L, K, 1) = EstBoots(L, K, Idx(I))
      Estimate(L, K, 2) = EstBoots(L, K, Idx(J))
      IF(Estimate(L, K, 0) <= Estimate(L, K, 1) .OR. &
         Estimate(L, K, 0) >= Estimate(L, K, 2)) &
        Estimate(L, K, 0) = EstBoots(L, K, Idx(M))
    END DO
  END DO       
END Subroutine FindCI95 
!
Subroutine Output
  USE AgeVariables
  IMPLICIT NONE
  INTEGER :: I, J, K, L
  Character(Len = 80) :: Line, Star, C
  CALL DATE_AND_TIME(Date2(2), Time2(2))  
  line = '-'
  Star = '*'
  DO I = 2, 80
    Line = Trim(Line) // '-'
    Star = Trim(Star) // '*'
  END DO
  IF(NumCI95Sample > 0) THEN
    J = 2
  ELSE
    J = 0
  END IF
  OPEN(10, File = Trim(FName(1)) // '.Ne')
  WRITE(10, '(A/T11,A/A)')Star(:70), 'AgeStructure Analysis Results for Dataset ' &
    // TRIM(DatasetName), Star(:70)
  WRITE(10, '(A)') &
    '  Time  Started: D=' // Date2(1)(7:8) // '/' // Date2(1)(5:6) // &
    '/' // Date2(1)(1:4) // ' T=' // Time2(1)(5:6) // &
    '/' // Time2(1)(3:4) // '/' // Time2(1)(1:2)
  WRITE(10, '(A)') &
    '  Time Finished: D=' // Date2(2)(7:8) // '/' // Date2(2)(5:6) // &
    '/' // Date2(2)(1:4) // ' T=' // Time2(2)(5:6) // &
    '/' // Time2(2)(3:4) // '/' // Time2(2)(1:2)
  WRITE(10, '(/A20/A/A20)')Line(:20), '1. Parameters', Line(:20)
  WRITE(10, '(/A30,2A15/A60)')'Parameter', 'Males', 'Females',Line
  WRITE(10, '(A30,2I15)')'Maximal Age', NumAgeClass(1 : 2)
  WRITE(10, '(A30,2I15)')'Minimal Parental Age', MinParentAge(1 : 2)
  IF(FamilySizeDist == 1) THEN
    WRITE(10, '(A30,2A15)')'Assumed Family Size Distr.', 'Poisson', 'Poisson'
  Else
    WRITE(10, '(A30,2A15)')'Assumed Family Size Distr.', &
      'NonPoisson', 'NonPoisson'
  END IF
  WRITE(10, '(A30, I15)')'Number bootstrapping samples', NumCI95Sample
  WRITE(10, '(A30, I15)')'Number of Loci', NumLoci
  WRITE(10, '(A30, T44, A2)')'Parentage confidence level', ConfidenceLevel
  WRITE(10, '(/A22/A/A22/)')Line, '2. Summary statistics', Line
  WRITE(10, '(A/)')'Number of individuals sampled in each sex-age class'
  WRITE(10, '(3A14/A42)')'AgeClass', 'Males', 'Females',Line
  DO L = 0, MaxNumAgeClass
    IF(ALL(L <= NumAgeClass(1 : 2))) THEN
      WRITE(10, '(3I14)') L, SexAgeSize0(L, 1 : 2)
    ELSE IF(L <= NumAgeClass(2)) THEN
      WRITE(10, '(I14,I28)') L, SexAgeSize0(L, 2)
    ELSE
      WRITE(10, '(2I14)') L, SexAgeSize0(L, 1)
    END IF     
  END DO 
  DO K = 1, 2
    IF(K == 1) THEN
      WRITE(10, '(//A/)')'Number of males & females of age' // &
        ' class j (column 1) whose paternity is assigned to age class i (row 1)'
    ELSE
      WRITE(10, '(//A/)')'Number of males & females of age' // &
        ' class j (column 1) whose maternity is assigned to age class i (row 1)'
    END IF
    WRITE(10, '(A10,999I6)')'AgeClass', (L, L = 1, NumAgeClass(K))
    WRITE(10, '(9999A)')('-', L = 1, 10 + 6 * NumAgeClass(K))
    DO L = 0, NumAgeClass(K) - 1
      WRITE(10, '(I10,999I6)')L, PODyadCounts0(L, 1 : NumAgeClass(K), K)
    END DO
  END DO
  DO K = 1, 2
    IF(K == 1) THEN
      WRITE(10, '(/A/)')'Number of male parents of age' // &
        ' class j (column 1) who each have i offspring assignments (row 1)'
    ELSE
      WRITE(10, '(/A/)')'Number of female parents of age' // &
        ' class j (column 1) who each have i offspring assignments (row 1)'
    END IF
    WRITE(10, '(A10,999I6)')'AgeClass', (L, L = 0, NumOff0Bound(K))
    WRITE(10, '(9999A)')('-', L = 1, 16 + 6 * NumOff0Bound(K))    
    DO L = 1, NumAgeClass(K)
      WRITE(10, '(I10,999I6)')L, NumOff0(0 : NumOff0Bound(K), L, K)
    END DO
  END DO  
  WRITE(10, '(/A20/A/A20)')Line(:30), '3. Estimates', Line(:30)
  WRITE(10, '(/A/)')'Estimates of generation effective size (Ne) ' // &
    '& generation intervals (GI)'
  WRITE(10, '(5A12)')' ', 'Ne', 'GI', 'Paternal-GI', 'Maternal-GI'
  WRITE(10, '(999A)')('-', L = 1, 12*5)
  WRITE(10, '(A12,I12,3F12.2)') 'Estimate', &
    NeEstimate(0), (GIEstimate(L, 0), L = 0, 2)
  IF(J > 0) THEN
    WRITE(10, '(A12,I12,3F12.2)') 'CI95L', &
      NeEstimate(1), (GIEstimate(L, 1), L = 0, 2)
    WRITE(10, '(A12,I12,3F12.2)') 'CI95U', &
      NeEstimate(2), (GIEstimate(L, 2), L = 0, 2)
  END IF    
  WRITE(10, '(//A)')'Estimates of variances and covariances of lifetime family size'
  WRITE(10, '(/A8,6A11)')' ', 'Var(mm)', 'Var(mf)', 'Cov(mm,mf)', &
    'Var(fm)', 'Var(ff)', 'Cov(fm,ff)'
  WRITE(10, '(9999A)')('-', L = 1, 74) 
  WRITE(10, '(A8, 6F11.2)')'Estimate', (Var(1 : 3, K, 0), K = 1, 2)
  IF(J > 0) THEN
    WRITE(10, '(A8, 6F11.2)')'CI95L', (Var(1 : 3, K, 1), K = 1, 2)
    WRITE(10, '(A8, 6F11.2)')'CI95U', (Var(1 : 3, K, 2), K = 1, 2)
  END IF
  DO K = 1, 4
    IF(K == 1) THEN
      WRITE(10, '(//A)')'Estimates of parameter P'
    ELSE IF(K == 2) THEN
      WRITE(10, '(//A)')'Estimates of parameter N'
    ELSE IF(K == 3) THEN
      IF(SamplingPrbKnown == 1) CYCLE
      WRITE(10, '(//A)')'Estimates of parameter S'
    ELSE 
      WRITE(10, '(//A)')'Estimates of parameter effective size of each age class'
    END IF
    IF(J > 0) THEN
      WRITE(10, '(/A/T23,A,T52,A)')Line(:68), 'Males', 'Females'
      WRITE(10, '(T11,A27,T40,A27)')Line(:27), Line(:27)
      WRITE(10, '(A8,6A10/A)')'AgeClass', 'Est', 'CI95L', &
        'CI95U', 'Est', 'CI95L', 'CI95U',Line(:68)
    ELSE
      WRITE(10, '(/A/A8, 2A12/A)')Line(:32), 'AgeClass', &
        'Est(Male)', 'Est(Female)',Line(:32)
    END IF   
    DO L = 1, MaxNumAgeClass
      IF(ALL(L <= NumAgeClass(1 : 2))) THEN
        IF(J > 0) THEN
          IF(Mod(K, 2) == 1) THEN
            C = '(I8, 6F10.4)'
          ELSE
            C = '(I8, 6F10.1)'
          END IF
        ELSE
          IF(Mod(K, 2) == 1) THEN
            C = '(I8, 2F12.4)'
          ELSE
            C = '(I8, 2F12.1)'
          END IF
        END IF          
        WRITE(10, Trim(C)) L, Estimate(L, K, 0 : J), &
          Estimate(NumAgeClass(1) + L, K, 0 : J)
      ELSE IF(L <= NumAgeClass(2)) THEN
        IF(J > 0) THEN
          IF(Mod(K, 2) == 1) THEN
            C = '(I8, F40.4, 2F10.4)'
          ELSE
            C = '(I8, F40.1, 2F10.1)'
          END IF
        ELSE
          IF(Mod(K, 2) == 1) THEN
            C = '(I8, F24.4)'
          ELSE
            C = '(I8, F24.1)'
          END IF          
        END IF  
        WRITE(10, Trim(C)) L, Estimate(NumAgeClass(1) + L, K, 0 : J) 
      ELSE
        IF(J > 0) THEN
          IF(Mod(K, 2) == 1) THEN
            C = '(I8, 3F10.4)'
          ELSE
            C = '(I8, 3F10.1)'
          END IF           
        ELSE
          IF(Mod(K, 2) == 1) THEN
            C = '(I8, 3F12.4)'
          ELSE
            C = '(I8, 3F12.1)'
          END IF            
        END IF
        WRITE(10, Trim(C)) L, Estimate(L, K, 0 : J)    
      END IF     
    END DO 
  END DO
  CLOSE(10)
END Subroutine Output
!
Subroutine ComputeNe(Ne, V, GISexEstimate, P, Size)
  Use AgeVariables
  IMPLICIT NONE
  INTEGER, INTENT(OUT) :: Ne
  REAL(DP), INTENT(OUT) :: V(3, 2), GISexEstimate(0 : 2)
  REAL(DP), INTENT(IN) :: P(NumAgeClass(0)), Size(NumAgeClass(0))
  REAL(DP) :: X, Q(NumAgeClass(0))
  INTEGER :: I, J, K, L
  DO I = 1, NumAgeClass(0)
    IF(I <= NumAgeClass(1)) THEN
      Q(I) = 2._DP * Sum(P(I : NumAgeClass(1)))
    ELSE
      Q(I) = 2._DP * Sum(P(I : NumAgeClass(0)))
    END IF
  END DO
  GISexEstimate(1) = Sum(Q(1 : NumAgeClass(1)))
  GISexEstimate(2) = Sum(Q(1 + NumAgeClass(1) : NumAgeClass(0)))
  GISexEstimate(0) = Sum(GISexEstimate(1 : 2)) * 0.5_DP
  DO J = 1, 2
    K = (J - 1) * NumAgeClass(1)
    X = 0._DP
    DO I = K + 2, K + NumAgeClass(J)
      X = X + Q(I) ** 2 * (1._DP / Size(I) - 1._DP / Size(I - 1))
    END DO
    X = X / 4._DP
    DO I = 1, 3
      IF(I == J) THEN
        V(I, J) = 1._DP + 4._DP * Size(K + 1) * X
      ELSE IF(I == 3 - J) THEN
        V(I, J) = Size(NumAgeClass(1) - K + 1) / Size(K + 1) * (1._DP + &
          4._DP * Size(NumAgeClass(1) - K + 1) * X)        
      ELSE 
        V(I, J) = 4._DP * Size(NumAgeClass(1) - K + 1) * X
      END IF
    END DO
  END DO
  X = 0._DP
  DO J = 1, 2
    IF(J == 1) THEN
      K = 0
    ELSE
      K = NumAgeClass(1)
    END IF
    X = X + 1._DP / Size(K + 1)
    DO I = 2, NumAgeClass(J)
      L = K + I
      X = X + Q(L) ** 2 * (1._DP / Size(L) - 1._DP / Size(L - 1))
    END DO
  END DO
  Ne = NINT(4._DP * GISexEstimate(0) / X)  
END Subroutine ComputeNe
!
Subroutine Simulate
  USE AgeVariables
  IMPLICIT NONE
  INTEGER :: I, J, K, L, M, K1, K2, K3, NG, NumIndiv
  REAL(DP) :: RM
!C Allocate and initialize variables  
  IF(.NOT. ALLOCATED(CP)) THEN
    Allocate(CP(MaxNumAgeClass, 2))
    DO J = 1, 2
      K = (J - 1) * NumAgeClass(1)
      CP(1, J) = ReproductionRate(K + 1) * 2
      DO I = 2, NumAgeClass(J) - 1
        CP(I, J) = CP(I - 1, J) + ReproductionRate(K + I) * 2
      END DO
      CP(NumAgeClass(J), J) = 1.1_DP
    END DO     
    Allocate(AgeClass(0 : MaxNumAgeClass, 2), GType(0 : MaxNumAgeClass, 2))
    DO J = 1, 2
      DO I = 0, NumAgeClass(J)
        Allocate(AgeClass(I, J)%M(0 : 6, AgeClassSize(I, J)), &
          GType(I, J)%G(2, NumLoci, AgeClassSize(I, J)))
      END DO
    END DO
    Allocate(Death(0 : MaxNumAgeClass, 2))
    DO J = 1, 2
      DO I = 0, NumAgeClass(J)
        Allocate(Death(I, J)%I(AgeClassSize(I, J)))
      END DO
      DO I = 1, AgeClassSize(0, J)
        Death(0, J)%I(I) = 0
      END DO  
      DO I = 1, AgeClassSize(NumAgeClass(J), J)
        Death(NumAgeClass(J), J)%I(I) = 1
      END DO  
    END DO  
    Allocate(AlleleFre(NumLoci, MaxObsNumAllele))
    DO I = 1, NumLoci
      AlleleFre(I, 1) = ObsAlleleFre(I, 1)
      DO J = 2, ObsNumAllele(I) - 1
        AlleleFre(I, J) = AlleleFre(I, J - 1) + ObsAlleleFre(I, J)
      END DO
      AlleleFre(I, ObsNumAllele(I)) = 1.1_DP
    END DO       
  END IF
!C Generate initial pedigree & genotypes of each individual in each age-class        
  NumIndiv = 0
  DO K1 = 1, 2
    DO I = 1, NumAgeClass(K1)
      DO J = 1, AgeClassSize(I, K1)
        NumIndiv = NumIndiv + 1
        AgeClass(I, K1)%M(0, J) = NumIndiv
        AgeClass(I, K1)%M(1 : 6, J) = 0
        DO K = 1, NumLoci
          DO L = 1, 2
            RM = RAN2()
            DO M = 1, ObsNumAllele(K)
              IF(RM <= AlleleFre(K, M)) THEN
                GType(I, K1)%G(L, K, J) = M
                EXIT
              END IF
            END DO
          END DO        
        END DO
      END DO
    END DO
  END DO
!C Iterate over time till all individuals have known grand parents
  NG = 0
  DO
!C Reproduction  
    DO K1 = 1, 2    
      DO I = 1, AgeClassSize(0, K1)
        NumIndiv = NumIndiv + 1
        AgeClass(0, K1)%M(0, I) = NumIndiv
        DO J = 1, 2      
          RM = RAN2()
          DO K = 1, NumAgeClass(J)
            IF(RM <= CP(K, J)) EXIT
          END DO          
          RM = RAN2()
          M = NINT(RM * (AgeClassSize(K, J) - 1)) + 1
          AgeClass(0, K1)%M(J, I) = AgeClass(K, J)%M(0, M)
          L = 2 * J        
          DO K2 = 1, 2
            AgeClass(0, K1)%M(L + K2, I) = AgeClass(K, J)%M(K2, M)  
          END DO          
          DO L = 1, NumLoci
            RM = RAN2()
            GType(0, K1)%G(J, L, I) = GType(K, J)%G(NINT(RM) + 1, L, M) 
          END DO
        END DO
      END DO
    END DO 
!C Check if grand parents or parents of all individuals known
    K1 = 1
    K2 = 2
    IF(NG == 0) THEN
      K = 0
      DO M = 1, 2
        DO I = NumAgeClass(M), 1, -1
          DO J = 1, AgeClassSize(I, M)
            IF(ANY(AgeClass(I, M)%M(K1 : K2, J) < 1)) THEN
              K = 1
              EXIT
            END IF
          END DO
          IF(K == 1) EXIT
        END DO
        IF(K == 1) EXIT
      END DO
      IF(K == 0) NG = 1
    ELSE
      NG = NG + 1
      IF(NG > 2 * MaxNumAgeClass) EXIT
    END IF
!C Death & Survival
    DO K1 = 1, 2  
      DO I = 0, NumAgeClass(K1) - 1
        IF(AgeClassSize(I, K1) > 2 * AgeClassSize(I + 1, K1)) THEN
          K2 = 1
          K3 = AgeClassSize(I + 1, K1)
        ELSE
          K2 = 0
          K3 = AgeClassSize(I, K1) - AgeClassSize(I + 1, K1)
        END IF
        DO J = 1, AgeClassSize(I, K1)
          Death(I, K1)%I(J) = K2
        END DO
        IF(K3 == 0) CYCLE          
        K2 = 1 - K2
        M = AgeClassSize(I, K1) - 1
        J = 0
        DO
          RM = RAN2()
          K = NINT(RM * M) + 1
          IF(Death(I, K1)%I(K) == K2) CYCLE
          Death(I, K1)%I(K) = K2
          J = J + 1
          IF(J == K3) EXIT
        END DO          
      END DO
    END DO
!C Transfer to older age class 
    DO M = 1, 2
      DO I = NumAgeClass(M), 1, -1
        K = 0
        DO J = 1, AgeClassSize(I - 1, M)
          IF(Death(I - 1, M)%I(J) == 1) CYCLE
          K = K + 1 
          DO K2 = 0, 6
            AgeClass(I, M)%M(K2, K) = AgeClass(I - 1, M)%M(K2, J)
          END DO                    
          DO K2 = 1, NumLoci
            DO K1 = 1, 2
              GType(I, M)%G(K1, K2, K) = GType(I - 1, M)%G(K1, K2, J) 
            END DO
          END DO          
        END DO
      END DO 
    END DO
  END DO    
!C select a random sample 
  K2 = 0
  DO
    RM = RAN2()
    DO M = 1, 2
      DO J = 0, NumAgeClass(M)
        IF(RM < SampleFre(J, M)) THEN
          RM = -1._DP
          EXIT
        END IF
      END DO
      IF(RM < 0._DP) EXIT
    END DO
    RM = RAN2()
    K = NINT(RM * (AgeClassSize(J, M) - 1)) + 1
    IF(AgeClass(J, M)%M(0, K) > 0) THEN
      AgeClass(J, M)%M(0, K) = -AgeClass(J, M)%M(0, K)
      K2 = K2 + 1
      IF(K2 == SmplSize) EXIT
    END IF
  END DO
  DO M = 1, 2
    DO J = 0, NumAgeClass(M)
      DO K = 1, AgeClassSize(J, M)
        AgeClass(J, M)%M(0, K) = -AgeClass(J, M)%M(0, K)
      END DO
    END DO
  END DO
END Subroutine Simulate
!
Subroutine FindEstimates(MaxAgeClassSize, NumIniPoint, NumAge, NN, P)
  USE AgeVariables
  USE PowellVariables
  USE FuncModule
  IMPLICIT NONE
  INTEGER, INTENT(IN) :: NumAge, NN, NumIniPoint, MaxAgeClassSize
  REAL(DP), INTENT(OUT) :: P(NN)
  REAL(DP) :: Ftol, Fret, RM, R, gammln
  REAL(DP), Allocatable :: XI(:, :), BestEst(:), Est(:, :)
  Logical :: FailQ
  INTEGER :: I, J, K
  EXTERNAL :: Powell, gammln
  FTol = 1.E-8_DP
  NDIM1 = NumAge - 1
  NDIM2 = NumAge
  IF(SamplingPrbKnown == 1) THEN
    NCOM = 2 * NumAge - 1
  ELSE
    NCOM = 3 * NumAge - 1
  END IF
  Allocate(XI(NCOM, NCOM), BestEst(0 : NCOM), &
    Est(0 : NCOM, NumIniPoint))  
  BestEst(0) = 1.E300_DP  
  R = Exp(Log(MaxAgeClassSize / DBLE(NumAge)) / DBLE(NumIniPoint))
  I = Max(Maxval(PODyadCounts(:, 0)), SampledCounts(0), UBound(NumOff, 1))
  IF(Allocated(LogFactorial)) THEN
    IF(UBound(LogFactorial, 1) >= I) THEN
      I = -1
    ELSE
      Deallocate(LogFactorial)
    END IF
  END IF
  IF(I > -1) THEN
    Allocate(LogFactorial(0 : I))
    LogFactorial(0 : 1) = 0._DP
    DO K = 2, I
      LogFactorial(K) = LogFactorial(K - 1) + Log(Dble(K))
    END DO
  END IF
  IF(Allocated(GammlnSave)) Deallocate(GammlnSave, gammlnSmpl)
  Allocate(GammlnSave(2, 0 : NumAge), gammlnSmpl(0 : NumAge))
  GammlnSave(1, :) = -1._DP
  J = 0
  DO I = 1, NumAge
    gammlnSmpl(I) = gammln(1._DP + SampledCounts(I))
    J = J + SampledCounts(I)
  END DO
  gammlnSmpl(0) = gammln(1._DP + J)  
  IF(Allocated(NumOffSum)) Deallocate(NumOffSum, NumOffSum2, SumNumOff)
  Allocate(NumOffSum(NumAge), NumOffSum2(NumAge), SumNumOff(NumAge))
  NumOffSum(:) = 0
  K = UBound(NumOff, 1)
  DO I = 1, NumAge
    DO J = 1, K
      NumOffSum(I) = NumOffSum(I) + J * NumOff(J, I)
    END DO
    NumOffSum2(I) = Sum(NumOff(0 : K, I))
    SumNumOff(I) = 0._DP
    DO J = 2, K
      SumNumOff(I) = SumNumOff(I) + LogFactorial(J) * NumOff(J, I)        
    END DO  
  END DO 
!Try NumIniPoints    
  DO K = NumIniPoint, 1, -1
    WRITE(*, Merge('(A  )', '(A,$)', Mod(K, 100) == 0 .AND. K < NumIniPoint)) '.'
    Fret = 1._DP
    DO
      DO I = 1, NumAge
        P(I) = RAN2()
      END DO
      RM = 1._DP / SUM(P(1 : NumAge))
      DO I = 1, NDIM1
        P(I) = P(I) * RM
      END DO
      RM = R ** K * DBLE(NumAge)
      IF(RM <= NumAge + 10) RM = RM + 1._DP       
      DO I = 1, NumAge
        P(NDIM1 + I) = RM - DBLE(I)
      END DO
      IF(SamplingPrbKnown /= 1) THEN
        DO I = 1, NumAge
          RM = RAN2()
          IF(RM < (SampledCounts(I) + .1_DP) / P(I + NDIM1)) &
             RM = (SampledCounts(I) + .1_DP) / P(I + NDIM1)
          P(NDIM1 + NDIM2 + I) = RM
        END DO
      ELSE
        DO I = 1, NumAge
          If(P(NDIM1 + I) * SamplingPrbPrt(I) <= SampledCounts(I)) THEN
            Fret = -1._DP
            EXIT
          END IF 
        END DO
        IF(Fret < 0._DP) EXIT
      END IF
      IF(P(NDIM1 + NumAge) < 1._DP .OR. &
        ANY(P(NDIM1 + NDIM2 + 1 : NCOM) > 1.0_DP)) THEN
        Fret = -1._DP
        EXIT
      END IF
      RM = Func(P)      
      IF(RM < -TinyValue) EXIT
    END DO    
    IF(Fret < 0._DP)THEN
      Est(0, K) = -TinyValue
      Est(1:, K) = 0._DP
      CYCLE
    END IF      
    XI(:, :) = 0.0_DP
    DO I = 1, NDIM1
      XI(I, I) = RAN2()
    END DO
    DO I = 1 + NDIM1, NDIM1 + NDIM2
      XI(I, I) = RAN2() * 100._DP + 1._DP
    END DO        
    DO I = 1 + NDIM1 + NDIM2, NCOM
      XI(I, I) = RAN2() * 0.1_DP
    END DO
    CALL Powell(P, XI, NCOM, NCOM, FTol, I, Fret, FailQ)    
    IF(FailQ) THEN
      Est(0, K) = -TinyValue
      Est(1:, K) = 0._DP
    ELSE
      Est(0, K) = Fret
      Est(1:, K) = P(:)      
      IF(Fret < BestEst(0)) THEN
        BestEst(0) = Fret
        BestEst(1 :) = P(:)
      END IF
    END IF
  END DO
  WRITE(*,*)  
  P(:) = BestEst(1 :)
  Fret = BestEst(0)
!C Mean estimates weighted by likelihood  
  RM = 0._DP
  DO I = 1, NumIniPoint
    Est(0, I) = exp(-Est(0, I) + Fret)
    IF(Est(0, I) < 0.99_DP) THEN
      Est(0, I) = 0._DP
    ELSE
      RM = RM + Est(0, I)
    END IF
  END DO
  P(:) = 0._DP
  DO I = 1, NumIniPoint
    Est(0, I) = Est(0, I) / RM
    IF(Est(0, I) <= 1.E-30_DP) CYCLE
    DO K = 1, NCOM
      P(K) = P(K) + Est(K, I) * Est(0, I)
    END DO
  END DO
END Subroutine FindEstimates
!
Subroutine EstimateByMarker(JR, NeByMarker, GIByMarker, VEst, NN, Est)
  USE AgeVariables
  USE PowellVariables  
  IMPLICIT NONE
  INTEGER, INTENT(IN) :: JR, NN
  INTEGER, INTENT(INOUT) :: NeByMarker
  REAL(DP), INTENT(INOUT) :: VEst(3, 2)
  REAL(DP), INTENT(INOUT) :: GIByMarker(0 : 2), Est(NN, 4)
  INTEGER :: IOS, I, J, K, L, N, K1, K2, K3, &
    NumAge(0 : 2), MaxNumAge, NumIndiv, NumIniPoint
  INTEGER, Pointer :: NPrt
  Character(Len = 20) :: CH
  Character(Len = 21), Allocatable :: IndivID(:, :)
  Character(Len = 21), Allocatable :: Individual(:)  
  INTEGER, Allocatable :: Age(:), Sex(:), IDPair(:, :)
  Type IntegerVector
    Integer, pointer :: I(:)
  END Type IntegerVector
  Type(IntegerVector), Allocatable :: SexAgeMember(:, :), ParentageAssigned(:)
  REAL(DP) :: V(2)
  REAL(DP), Allocatable :: P(:), Size(:)
  EXTERNAL :: StopOnError, FindEstimates
  IF(JR == 0) THEN 
    NumIniPoint = 500
    WRITE(*,'(//T20,A/T20,A/T20,A//)')&
      '-----------------------------------------', &
      '      Estimating from original data', &
      '-----------------------------------------'
  ELSE
    NumIniPoint = 100
    WRITE(CH, '(I6)') JR
    WRITE(*,'(//T20,A/T20,A/T20,A//)')&
      '-----------------------------------------------', &
      '  Estimating from simulated data: replicate=' // AdjustL(CH), &
      '-----------------------------------------------'    
  END IF
  Allocate(Individual(SmplSize), Sex(SmplSize), Age(SmplSize))
  IF(SamplingPrbKnown == 1) THEN
    N = 16
  ELSE
    N = 14
  END IF
  OPEN(UNIT=10,FILE='AgeSexGtype.dat',IOSTAT=IOS,ERR=1,STATUS='OLD')
  DO I = 1, N
    READ(10, *) CH
  END DO
  DO I = 1, SmplSize
    READ(10, *)Individual(I), Sex(I), Age(I)
    Individual(I) = Adjustl(Individual(I))
    Individual(I) = Individual(I)(1 : 20)
  END DO
  CLOSE(10)
  NumAge(:) = NumAgeClass(:)
  MaxNumAge = Maxval(NumAge(1 : 2))
  NumAge(0) = Sum(NumAge(1 : 2))
  MaxNumAgeClass = MaxNumAge
  IF(.Not. Allocated(SexAgeSize)) Allocate(SexAgeSize(0 : MaxNumAge, 2)) 
  Allocate(SexAgeMember(0 : MaxNumAge, 2))
  IF(.NOT. Allocated(Factor)) Allocate(Factor(MaxNumAge))  
  SexAgeSize(:, :) = 0
  DO I = 1, SmplSize
    SexAgeSize(Age(I), Sex(I)) = SexAgeSize(Age(I), Sex(I)) + 1
  END DO
  DO I = 1, 2
    DO J = 0, NumAge(I)
      Allocate(SexAgeMember(J, I)%I(SexAgeSize(J, I)))
    END DO
  END DO
  SexAgeSize(:, :) = 0
  DO I = 1, SmplSize
    SexAgeSize(Age(I), Sex(I)) = SexAgeSize(Age(I), Sex(I)) + 1
    SexAgeMember(Age(I), Sex(I))%I(SexAgeSize(Age(I), Sex(I))) = I
  END DO 
  IF(JR == 0) THEN
    Allocate(SexAgeSize0(0 : MaxNumAge, 2), &
      PODyadCounts0(0 : MaxNumAge, MaxNumAge, 2), &
      NumOff0(0 : SmplSize, MaxNumAge, 2))
    SexAgeSize0 = SexAgeSize
  END IF
  CALL RemoveFile(Trim(FNAME(1)) // '.PairwisePaternity')
  CALL RemoveFile(Trim(FNAME(1)) // '.PairwiseMaternity')
  IF(CrrtOS == 1) THEN
    CALL System('POAssignment')
  ELSE
    CALL System('./POAssignment')
  END IF
  IF(JR == 0) THEN      
    CALL CopyFile(Trim(FNAME(1)) // '.PairwisePaternity', Trim(FNAME(1)) // '.Paternity')
    CALL CopyFile(Trim(FNAME(1)) // '.PairwiseMaternity', Trim(FNAME(1)) // '.Maternity')     
  END IF
  Allocate(IndivID(2, SmplSize), IDPair(2, SmplSize))
  IF(Allocated(SampledCounts)) Deallocate(SampledCounts)
  Allocate(SampledCounts(0 : MaxNumAge))
  DO N = 1, 2
!C Find number of individuals sampled in each age class of sex j
    DO I = 1, NumAge(N)                            !offspring age
      SampledCounts(I) = SexAgeSize(I, N)
    END DO    
    SampledCounts(0) = Sum(SampledCounts(1 : NumAge(N)))    
    IF(N == 1) THEN
      FName(2) = Trim(FName(1)) // '.PairwisePaternity'
      WRITE(*,*)'Estimating paternal parameters from paternity assigments'
    ELSE
      FName(2) = Trim(FName(1)) // '.PairwiseMaternity'
      WRITE(*,*)'Estimating maternal parameters from maternity assigments'      
    END IF
    OPEN(UNIT=10,FILE=Trim(FName(2)),IOSTAT=IOS,ERR=1,STATUS='OLD')
    READ(10,*,IOSTAT=IOS,ERR=9,End=9)CH
    NumIndiv = 0
    DO 
      NumIndiv = NumIndiv + 1
      READ(10,*,IOSTAT=IOS,ERR=9,End=9) IndivID(1 : 2, NumIndiv), CH
      CH = AdjustL(CH)
      IF(CH(1 : 2) < ConfidenceLevel .AND. CH(1 : 3) /= '100') THEN
        NumIndiv = NumIndiv - 1
        CYCLE
      END IF
      DO I = 1, 2
        IndivID(I, NumIndiv) = AdjustL(IndivID(I, NumIndiv))
        DO J = 1, SmplSize
          IF(IndivID(I, NumIndiv) == Individual(J)) THEN
            IDPair(I, NumIndiv) = J
            EXIT
          END IF
        END DO
      END DO
    END DO
9   CLOSE(10)
    NumIndiv = NumIndiv - 1
!C Find the count of events that the parentage of an offspring of age i is assigned 
!    to an individual of age j (i < j) 
    IF(Allocated(PODyadCounts)) DeAllocate(PODyadCounts)
    Allocate(PODyadCounts(0 : NumAge(N) - 1, 0 : NumAge(N)))
    PODyadCounts(:, :) = 0
    DO I = 1, NumIndiv
      K2 = Age(IDPair(1, I))
      IF(K2 > NumAge(N) - 1) CYCLE
      K3 = Age(IDPair(2, I))
      PODyadCounts(K2, K3) = PODyadCounts(K2, K3) + 1
    END DO
    DO K1 = 0, NumAge(N) - 1
      PODyadCounts(K1, 0) = Sum(SexAgeSize(K1, 1 : 2))    !# offspring sampled
    END DO 
    IF(JR == 0) THEN
      PODyadCounts0(0 : NumAge(N) - 1, 1 : NumAge(N), N) = &
        PODyadCounts(0 : NumAge(N) - 1, 1 : NumAge(N))
    END IF
!C Find the freq. that 2 offspring, whose parentage are both assigned to parents 
!    of same age, are from the same parent   
    IF(Allocated(ParentageAssigned)) Deallocate(ParentageAssigned)
    Allocate(ParentageAssigned(NumAge(N)))
    DO K = 1, NumAge(N) 
      Allocate(ParentageAssigned(K)%I(SexAgeSize(K, N)))
      DO I = 1, SexAgeSize(K, N)
        ParentageAssigned(K)%I(I) = 0
      END DO
    END DO
    DO I = 1, NumIndiv   
      J = IDPair(2, I)
      DO K = 1, SexAgeSize(Age(J), Sex(J))
        IF(SexAgeMember(Age(J), Sex(J))%I(K) == J) then
          ParentageAssigned(Age(J))%I(K) = ParentageAssigned(Age(J))%I(K) + 1
          EXIT
        END IF
      END DO
    END DO  
    K1 = 0
    DO L = 1, NumAge(N)
      DO K = 1, SexAgeSize(L, N)
        IF(K1 < ParentageAssigned(L)%I(K)) K1 = ParentageAssigned(L)%I(K)
      END DO
    END DO
    IF(Allocated(NumOff)) DeAllocate(NumOff)
    Allocate(NumOff(0 : K1, NumAge(N)))
    DO L = 1, NumAge(N)
      NumOff(0 : K1, L) = 0
      DO K = 1, SexAgeSize(L, N)
        NPrt => ParentageAssigned(L)%I(K)
        NumOff(NPrt, L) = NumOff(NPrt, L) + 1
      END DO
    END DO
    IF(JR == 0) THEN
      NumOff0Bound(N) = K1
      NumOff0(0 : K1, 1 : NumAge(N), N) = NumOff(0 : K1, 1 : NumAge(N))
    END IF    
    IF(JR == 0) THEN
      IF(FamilySizeDist /= 1) THEN !non-Poisson distribution
        DO L = 1, NumAge(N)
          V(:) = 0._DP
          DO K = 1, SexAgeSize(L, N)
            NPrt => ParentageAssigned(L)%I(K)
            V(1) = V(1) + NPrt * (NPrt - 1)
            V(2) = V(2) + NPrt 
          END DO
          IF(V(1) > .1_DP) THEN
            V(2) = V(2) * (V(2) - 1)
            V(1) = V(1) / V(2)
            Factor(L) = (1._DP / SexAgeSize(L, N)) / V(1)
          ELSE
            Factor(L) = 1._DP
          END IF                 
        END DO
      ELSE
        Factor(:) = 1._DP
      END IF
    END IF
!C Estimate reproduction rate and GI
    IF(Allocated(P)) Deallocate(P)
    IF(SamplingPrbKnown /= 1) THEN
      K = 3 * NumAge(N) - 1
    ELSE
      K = 2 * NumAge(N) - 1
      SamplingPrbPrt => SamplingPrbPrior(N)%R
    END IF
    Allocate(P(K))
    IF(JR == 0) THEN
      L = 10000
    ELSE
      L = Maxval(AgeClassSize(1 : NumAgeClass(N), N)) * 100
      IF(L > 1.E5_DP) L = 100000
    END IF
    CALL FindEstimates(L, NumIniPoint, NumAge(N), K, P)
    P(1 : NDIM1) = P(1 : NDIM1) * 0.5_DP
    K = (N - 1) * NumAge(1)  
    Est(K + 1 : K + NDIM1, 1) = P(1 : NDIM1)
    Est(K + NDIM1 + 1, 1) = 0.5_DP - Sum(P(1 : NDIM1))    
    Est(K + 1 : K + NDIM2, 2) = P(NDIM1 + 1 : NDIM1 + NDIM2)
    IF(SamplingPrbKnown /= 1) Est(K + 1 : K + NDIM2, 3) = &
      P(1 + NDIM1 + NDIM2 : NCOM)
    Est(K + 1 : K + NDIM2, 4) = P(NDIM1 + 1 : NDIM1 + NDIM2) &
      * Factor(1 : NumAge(N))
  END DO
  IF(Allocated(P)) Deallocate(P)
  Allocate(P(NumAgeClass(0)), Size(NumAgeClass(0)))
  P(:) = Est(:, 1)
  Size(:) = Est(:, 4)
  CALL ComputeNe(NeByMarker, VEst, GIByMarker, P, Size)
  RETURN
1 CALL StopOnError('Error in reading file ID_Sex_Age.dat or other files')
END Subroutine EstimateByMarker 
!
Subroutine WriteColonyData
  USE AgeVariables
  IMPLICIT NONE
  INTEGER :: I, J, K, L, M, N, LL
  Character(Len = 10) :: C, CH
  Character(Len = 30) :: IndivID(SmplSize)
  Character(Len = 30000) :: String
  Integer, Allocatable :: Sex(:), Age(:), Idx(:), Jdx(:)
  OPEN(10, File='AgeSexGtype.dat')
  WRITE(10, '(A,T30,A)')TRIM(DatasetName), ' ! C, Dataset name, Length<51' 
  WRITE(10, '(A,T30,A)')TRIM(FNAME(1)), ' ! C, Main output file name, Length<21'
  WRITE(10, '(I10,T30,A)')SmplSize, ' ! I, Number of offspring in the sample'
  WRITE(10, '(2I5,T30,A)')NumAgeClass(1 : 2), ' ! Max age for males & females'
  WRITE(10, '(2I5,T30,A)')MinParentAge(1 : 2), ' ! Minimum age for reproduction'
  WRITE(10, '(I10,T30,A)')FamilySizeDist, ' !=1/0 for Poisson/Other dist'
  WRITE(10, '(I10,T30,A)')SamplingPrbKnown, &
    ' !=1/0 =sampling proportion known for each age class'
  IF(SamplingPrbKnown == 1) THEN
    DO K = 1, 2
      WRITE(10, '(999F10.5)')(SamplingPrbPrior(K)%R(J), J = 1, NumAgeClass(K))
    END DO
  END IF
  WRITE(10, '(I10,T30,A)')NumCI95Sample, ' !=Number of CI95 samples'
  WRITE(10, '(I10,T30,A)')NumLoci, ' ! I, Number of loci'
  WRITE(10, '(A2,T30,A)')ConfidenceLevel, ' ! Confidence level for parentage assignment'
  WRITE(10, '(999A)')(TRIM(AdjustL(MarkerID(I))) // ' ', I = 1, Numloci)
  WRITE(10, '(999I2)')(Dominant(I), I = 1, Numloci)  
  WRITE(10, '(999F8.4)')DropRate(1 : NumLoci) !Drop rate for each locus
  WRITE(10, '(999F8.4)')OtherRate(1 : NumLoci) !OtherErrorRate for each locus
  I = MaxIntNumber
  DO M = 1, 2
    DO J = 0, NumAgeClass(M)
      DO K = 1, AgeClassSize(J, M)  
        IF(AgeClass(J, M)%M(0, K) < 1) CYCLE 
        DO L = 0, 6
          IF(AgeClass(J, M)%M(L, K) < I) I = AgeClass(J, M)%M(L, K)
        END DO
      END DO
    END DO
  END DO
  I = I - 1
  N = 0
  DO M = 1, 2
    DO J = 0, NumAgeClass(M)
      DO K = 1, AgeClassSize(J, M)  
        IF(AgeClass(J, M)%M(0, K) < 1) CYCLE 
        DO L = 0, 6
          AgeClass(J, M)%M(L, K) = AgeClass(J, M)%M(L, K) - I
          IF(AgeClass(J, M)%M(L, K) > N) N = AgeClass(J, M)%M(L, K)
        END DO
      END DO
    END DO
  END DO
  Allocate(Jdx(N))  
  Jdx(:) = 0
  N = 0
  DO M = 1, 2
    DO J = 0, NumAgeClass(M)
      DO K = 1, AgeClassSize(J, M)  
        IF(AgeClass(J, M)%M(0, K) < 1) CYCLE 
        DO L = 0, 6
          IF(Jdx(AgeClass(J, M)%M(L, K)) /= 0) CYCLE
          N = N + 1
          Jdx(AgeClass(J, M)%M(L, K)) = N
        END DO
      END DO
    END DO
  END DO
  DO M = 1, 2
    DO J = 0, NumAgeClass(M)
      DO K = 1, AgeClassSize(J, M)  
        IF(AgeClass(J, M)%M(0, K) < 1) CYCLE 
        DO L = 0, 6
          AgeClass(J, M)%M(L, K) = Jdx(AgeClass(J, M)%M(L, K))
        END DO
      END DO
    END DO
  END DO  
  Allocate(Sex(SmplSize), Age(SmplSize))
  Allocate(Idx(NumLoci))   
  DO L = 1, NumLoci
    IF(Dominant(L) == 1) THEN
      Idx(L) = 1
    ELSE
      Idx(L) = 2
    END IF
  END DO  
  I = 0
  DO N = 1, 2
    DO J = 0, NumAgeClass(N)
      WRITE(C, '(I6)')J
      IF(N == 1) THEN
        C = 'M' // AdjustL(C) 
      ELSE
        C = 'F' // AdjustL(C) 
      END IF      
      M = 0
      DO K = 1, AgeClassSize(J, N)  
        IF(AgeClass(J, N)%M(0, K) < 1) CYCLE
        I = I + 1
        Sex(I) = N
        Age(I) = J
        WRITE(CH, '(I6)')AgeClass(J, N)%M(0, K)
        IndivID(I) = Trim(C) // 'O' // AdjustL(CH)
        WRITE(CH, '(I6)')AgeClass(J, N)%M(1, K)
        IndivID(I) = Trim(IndivID(I)) // 'P' // AdjustL(CH)
        WRITE(CH, '(I6)')AgeClass(J, N)%M(2, K)
        IndivID(I) = Trim(IndivID(I)) // 'P' // AdjustL(CH)
        DO L = 3, 6
          WRITE(CH, '(I6)')AgeClass(J, N)%M(L, K)       
          IndivID(I) = Trim(IndivID(I)) // 'G' // AdjustL(CH)
        END DO  
        DO L = 1, NumLoci
          IF(Dominant(L) == 1) THEN
            IF(ANY(GType(J, N)%G(1 : 2, L, K) == 1)) &
              GType(J, N)%G(1, L, K) = 1
          END IF
        END DO
        String = IndivID(I)
        WRITE(CH, '(I6)')Sex(I)
        String = Trim(AdjustL(String)) // ',' // Trim(AdjustL(CH)) 
        WRITE(CH, '(I6)')Age(I)
        String = Trim(String) // ',' // Trim(AdjustL(CH))
        DO L = 1, NumLoci
          DO LL = 1, 2 - Dominant(L)
            WRITE(CH, '(I8)')GType(J, N)%G(LL, L, K)
            String = Trim(String) // ',' // Trim(AdjustL(CH))
          END DO 
        END DO
        WRITE(10,'(A)')TRIM(String)
        M = M + 1
      END DO
    END DO
  END DO
  WRITE(10, '(2F8.4,T30,A)')PropPrntSampled(1 : 2), '!sampling proportion for males & females'
  WRITE(10, '(I10,T30,A)')0, '!number of known paternity'
  WRITE(10, '(I10,T30,A)')0, '!number of known maternity'
  WRITE(10, '(I10,T30,A)')0, '!number of excluded paternity'
  WRITE(10, '(I10,T30,A)')0, '!number of excluded maternity'
  CLOSE(10)
END Subroutine WriteColonyData  
!
FUNCTION ran2()
  USE AgeVariables, ONLY : IDUM, DP
  INTEGER IM1,IM2,IMM1,IA1,IA2,IQ1,IQ2,IR1,IR2,NTAB,NDIV
  REAL(DP) :: ran2,AM,EPS,RNMX
  PARAMETER (IM1=2147483563,IM2=2147483399,AM=1._DP/IM1,IMM1=IM1-1, &
    IA1=40014,IA2=40692,IQ1=53668,IQ2=52774,IR1=12211,IR2=3791, &
    NTAB=32,NDIV=1+IMM1/NTAB,EPS=1.2e-7_DP,RNMX=1._DP-EPS)
  INTEGER idum2,j,k,iv(NTAB),iy
  SAVE iv,iy,idum2
  DATA idum2/123456789/, iv/NTAB*0/, iy/0/
  if (idum <= 0) then
    idum=max(-idum,1)
    idum2=idum
    do j=NTAB+8,1,-1
      k=idum/IQ1
      idum=IA1*(idum-k*IQ1)-k*IR1
      if (idum < 0) idum=idum+IM1
      if (j <= NTAB) iv(j)=idum
    end do
    iy=iv(1)
  endif
  k=idum/IQ1
  idum=IA1*(idum-k*IQ1)-k*IR1
  if (idum < 0) idum=idum+IM1
  k=idum2/IQ2
  idum2=IA2*(idum2-k*IQ2)-k*IR2
  if (idum2 < 0) idum2=idum2+IM2
  j=1+iy/NDIV
  iy=iv(j)-idum2
  iv(j)=idum
  if(iy < 1)iy=iy+IMM1
  ran2=min(AM*iy,RNMX)
  return
END FUNCTION ran2
!
SUBROUTINE powell(p,xi,n,np,ftol,iter,fret, FailQ)
  Use AgeVariables, ONLY : DP
  Use PowellVariables
  USE FuncModule
  IMPLICIT NONE
  Logical :: FailQ
  INTEGER iter,n,np,ITMAX
  REAL(DP) :: fret,ftol,p(np),xi(np,np),TINY
  PARAMETER (ITMAX=200,TINY=1.e-25_DP)
!C USES func,linmin
  INTEGER :: i,ibig,j
  REAL(DP) :: del,fp,fptt,t,pt(N),ptt(n),xit(n)
  FailQ = .FALSE.
  IF(Allocated(PCom)) Deallocate(PCom, XICom)
  Allocate(pcom(n),xicom(n))
  fret=func(p)
  do j=1,n
    pt(j)=p(j)
  end do
  iter=0
  DO
    iter=iter+1
    fp=fret
    ibig=0
    del=0._DP
    do i=1,n
      do j=1,n
        xit(j)=xi(j,i)
      end do
      fptt=fret
      call linmin(p,xit,n,fret, FailQ)
      IF(FailQ) RETURN
      if(fptt-fret > del)then
        del=fptt-fret
        ibig=i
      endif
    end do
    if(2._DP*(fp-fret) <= ftol*(abs(fp)+abs(fret))+TINY)return
    if(iter == ITMAX) THEN
      FailQ = .true.
      RETURN
    end if
    do j=1,n
      ptt(j)=2._DP*p(j)-pt(j)
      xit(j)=p(j)-pt(j)
      pt(j)=p(j)
    end do
    fptt=func(ptt)
    if(fptt >= fp) CYCLE
    t=2._DP*(fp-2._DP*fret+fptt)*(fp-fret-del)**2-del*(fp-fptt)**2
    if(t >= 0._DP) CYCLE
    call linmin(p,xit,n,fret, FailQ)
    IF(FailQ) RETURN
    do j=1,n
      xi(j,ibig)=xi(j,n)
      xi(j,n)=xit(j)
    end do
  end do
END SUBROUTINE powell
!
SUBROUTINE linmin(p,xi,n,fret, FailQ)
  Use AgeVariables, ONLY : DP
  Use PowellVariables
  IMPLICIT NONE
  Logical :: FailQ
  INTEGER :: n
  REAL(DP) :: fret,p(n),xi(n),TOL
  PARAMETER (TOL=1.e-8_DP)
!C USES brent,f1dim,mnbrak
  INTEGER :: j
  REAL(DP) :: ax,bx,fa,fb,fx,xmin,xx,brent
  REAL(DP) :: f1dim
  EXTERNAL :: f1dim
  do j=1,n
    pcom(j)=p(j)
    xicom(j)=xi(j)
  end do
  ax=0._DP
  xx=1._DP
  call mnbrak(ax,xx,bx,fa,fx,fb,f1dim, FailQ)
  IF(.NOT. FailQ) fret=brent(ax,xx,bx,f1dim,TOL,xmin, FailQ)
  IF(FailQ) RETURN
  do j=1,n
    xi(j)=xmin*xi(j)
    p(j)=p(j)+xi(j)
  end do
END SUBROUTINE linmin
!
FUNCTION f1dim(x)
  Use AgeVariables, ONLY : DP, I8R
  Use PowellVariables
  USE FuncModule
  IMPLICIT NONE
  REAL(DP) :: f1dim,x
  REAL(I8R) :: XLimit
  INTEGER :: j
  REAL(DP) :: xt(ncom)
  do j=1,ncom
    IF(ABS(xicom(j)) <= 0._DP) THEN
      xt(j)= pcom(j)
    ELSE    
      XLimit = pcom(j)
      XLimit = XLimit + x*xicom(j)
      IF(XLIMIT < -9.E307_DP .OR. XLIMIT > 9.E307_DP) THEN
        f1dim = 1.E300_DP
        RETURN
      END IF
      xt(j)=DBLE(XLimit)
    END IF
  END DO
  f1dim=func(xt)
END FUNCTION f1dim
!
SUBROUTINE mnbrak(ax,bx,cx,fa,fb,fc,func, FailQ)
  Use AgeVariables, ONLY : DP, I8R
  IMPLICIT NONE
  Logical, INTENT(OUT) :: FailQ
  REAL(DP) :: ax,bx,cx,fa,fb,fc,func,GOLD,GLIMIT,TINY
  EXTERNAL func
  PARAMETER (GOLD=1.618034_DP, GLIMIT=100._DP, TINY=1.e-20_DP)
  REAL(DP) :: dum,fu,q,r,u,ulim
  REAL(I8R) :: ULIMIT
  FailQ = .FALSE.
  fa=func(ax)
  fb=func(bx)
  if(fb > fa)then
    dum=ax
    ax=bx
    bx=dum
    dum=fb
    fb=fa
    fa=dum
  endif
  cx=bx+GOLD*(bx-ax)
  fc=func(cx)
1 if(fb >= fc)then
    r=(bx-ax)*(fb-fc)
    q=(bx-cx)*(fb-fa)
    u=bx-((bx-cx)*q-(bx-ax)*r)/(2._DP*sign(max(abs(q-r),TINY),q-r))
    ULIMIT = bx+GLIMIT*(cx-bx)
    IF(ULIMIT < -9.E307_DP .OR. ULIMIT > 9.E307_DP) THEN
      FailQ = .TRUE.
      RETURN
    END IF
    ulim=bx+GLIMIT*(cx-bx)
    if((bx-u)*(u-cx) > 0._DP)then
      fu=func(u)
      if(fu < fc)then
        ax=bx
        fa=fb
        bx=u
        fb=fu
        return
      else if(fu > fb)then
        cx=u
        fc=fu
        return
      endif
      u=cx+GOLD*(cx-bx)
      fu=func(u)
    else if((cx-u)*(u-ulim) > 0._DP)then
      fu=func(u)
      if(fu < fc)then
        bx=cx
        cx=u
        u=cx+GOLD*(cx-bx)
        fb=fc
        fc=fu
        fu=func(u)
      endif
    else if((u-ulim)*(ulim-cx) >= 0._DP)then
      u=ulim
      fu=func(u)
    else
      u=cx+GOLD*(cx-bx)
      fu=func(u)
    endif
    ax=bx
    bx=cx
    cx=u
    fa=fb
    fb=fc
    fc=fu
    goto 1
  endif
END SUBROUTINE mnbrak
!
FUNCTION brent(ax,bx,cx,f,tol,xmin, FailQ)
  Use AgeVariables, ONLY : DP, I8R
  IMPLICIT NONE
  Logical :: FailQ
  INTEGER :: ITMAX
  REAL(DP) :: brent,ax,bx,cx,tol,xmin,f,CGOLD,ZEPS,SmallReal
  EXTERNAL f
  PARAMETER (ITMAX=100,CGOLD=.3819660_DP, &
    ZEPS=1.0e-10_DP,SmallReal=1.0E-30_DP)
  INTEGER :: iter
  REAL(DP) :: a,b,d,e,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm
  REAL(I8R) :: PLimit
  a=min(ax,cx)
  b=max(ax,cx)
  v=bx
  w=v
  x=v
  e=0._DP
  fx=f(x)
  fv=fx
  fw=fx
  do iter=1,ITMAX
    xm=0.5_DP*(a+b)
    tol1=tol*abs(x)+ZEPS
    tol2=2._DP*tol1
    if(abs(x-xm) <= (tol2-.5_DP*(b-a))) goto 3
    if(abs(e) > tol1) then
      PLimit = x-w
      PLimit = PLimit*(fx-fv)
      IF(PLimit < -9.E307_DP .OR. PLimit > 9.E307_DP) THEN
        FailQ = .TRUE.
        RETURN
      END IF
      PLimit = x-v
      PLimit = PLimit*(fx-fw)
      IF(PLimit < -9.E307_DP .OR. PLimit > 9.E307_DP) THEN
        FailQ = .TRUE.
        RETURN
      END IF       
      r=(x-w)*(fx-fv)
      q=(x-v)*(fx-fw)  
      PLimit = X - V
      PLimit = PLimit * q-(x-w)*r  
      IF(PLimit < -9.E307_DP .OR. PLimit > 9.E307_DP) THEN
        FailQ = .TRUE.
        RETURN
      END IF
      p=(x-v)*q-(x-w)*r
      q=2._DP*(q-r)
      if(q > 0._DP) p=-p
      q=abs(q)
      etemp=e
      e=d
      if(abs(p) >= abs(.5_DP*q*etemp).or.p <= q*(a-x).or.p >= q*(b-x)) goto 1
      d=p/q
      u=x+d
      if(u-a < tol2 .or. b-u < tol2) d=sign(tol1,xm-x)
      goto 2
    endif
1   if(x >= xm) then
      e=a-x
    else
      e=b-x
    endif
    d=CGOLD*e
2   if(abs(d) >= tol1) then
      u=x+d
    else
      u=x+sign(tol1,d)
    endif
    fu=f(u)
    if(fu <= fx) then
      if(u >= x) then
        a=x
      else
        b=x
      endif
      v=w
      fv=fw
      w=x
      fw=fx
      x=u
      fx=fu
    else
      if(u < x) then
        a=u
      else
        b=u
      endif
      if(fu <= fw .or. ABS(w - x) < SmallReal) then
        v=w
        fv=fw
        w=u
        fw=fu
      else if(fu <= fv .or. ABS(v - x) < SmallReal .or. &
              ABS(v - w) < SmallReal) then
        v=u
        fv=fu
      endif
    endif
  END DO
  FailQ = .true.
3 xmin=x
  brent=fx
  return
END FUNCTION brent
!
SUBROUTINE indexx(n,arr,indx)
  IMPLICIT NONE
  INTEGER, INTENT(IN) :: n
  DOUBLE PRECISION, INTENT(IN) :: arr(n)
  INTEGER, INTENT(OUT) :: indx(n)
  INTEGER M,NSTACK
  DOUBLE PRECISION a
  PARAMETER (M=7,NSTACK=50)
  INTEGER i,indxt,ir,itemp,j,jstack,k,l,istack(NSTACK)
  EXTERNAL :: StopOnError
  FORALL(j=1 : n) indx(j)=j
  jstack=0
  l=1
  ir=n
1 if(ir-l < M)then
    DO j=l+1,ir
      indxt=indx(j)
      a=arr(indxt)
      DO i=j-1,l,-1
        if(arr(indx(i)) <= a)goto 2
        indx(i+1)=indx(i)
      END DO
      i=l-1
2     indx(i+1)=indxt
    END DO
    if(jstack == 0)return
    ir=istack(jstack)
    l=istack(jstack-1)
    jstack=jstack-2
  else
    k=(l+ir)/2
    itemp=indx(k)
    indx(k)=indx(l+1)
    indx(l+1)=itemp
    if(arr(indx(l)) > arr(indx(ir)))then
      itemp=indx(l)
      indx(l)=indx(ir)
      indx(ir)=itemp
    endif
    if(arr(indx(l+1)) > arr(indx(ir)))then
      itemp=indx(l+1)
      indx(l+1)=indx(ir)
      indx(ir)=itemp
    endif
    if(arr(indx(l)) > arr(indx(l+1)))then
      itemp=indx(l)
      indx(l)=indx(l+1)
      indx(l+1)=itemp
    endif
    i=l+1
    j=ir
    indxt=indx(l+1)
    a=arr(indxt)
3   continue
    i=i+1
    if(arr(indx(i)) < a)goto 3
4   continue
    j=j-1
    if(arr(indx(j)) > a)goto 4
    if(j < i)goto 5
    itemp=indx(i)
    indx(i)=indx(j)
    indx(j)=itemp
    goto 3
5   indx(l+1)=indx(j)
    indx(j)=indxt
    jstack=jstack+2
    if(jstack > NSTACK) THEN
      CALL StopOnError('StackError in subroutine indexx')
    end if
    if(ir-i+1 >= j-l)then
      istack(jstack)=ir
      istack(jstack-1)=i
      ir=j-1
    else
      istack(jstack)=j-1
      istack(jstack-1)=l
      l=i
    endif
  endif
  goto 1
END SUBROUTINE indexx    
!
SUBROUTINE indexCh(n, ChLength, indx)
  Use AgeVariables
  IMPLICIT NONE
  INTEGER, INTENT(IN) :: n, ChLength
  INTEGER, INTENT(OUT) :: indx(n)
  Character(Len = ChLength) :: a
  INTEGER, PARAMETER :: M=7,NSTACK=50
  INTEGER :: i,indxt,ir,itemp,j,jstack,k,l,istack(NSTACK)
  FORALL(j=1 : n) indx(j)=j
  jstack=0
  l=1
  ir=n
1 if(ir-l < M)then
    DO j=l+1,ir
      indxt=indx(j)
      a=SortCharacter(indxt)
      DO i=j-1,l,-1
        if(SortCharacter(indx(i)) <= a)goto 2
        indx(i+1)=indx(i)
      END DO
      i=l-1
2     indx(i+1)=indxt
    END DO
    if(jstack == 0)return
    ir=istack(jstack)
    l=istack(jstack-1)
    jstack=jstack-2
  else
    k=(l+ir)/2
    itemp=indx(k)
    indx(k)=indx(l+1)
    indx(l+1)=itemp
    if(SortCharacter(indx(l)) > SortCharacter(indx(ir)))then
      itemp=indx(l)
      indx(l)=indx(ir)
      indx(ir)=itemp
    endif
    if(SortCharacter(indx(l+1)) > SortCharacter(indx(ir)))then
      itemp=indx(l+1)
      indx(l+1)=indx(ir)
      indx(ir)=itemp
    endif
    if(SortCharacter(indx(l)) > SortCharacter(indx(l+1)))then
      itemp=indx(l)
      indx(l)=indx(l+1)
      indx(l+1)=itemp
    endif
    i=l+1
    j=ir
    indxt=indx(l+1)
    a=SortCharacter(indxt)
3   continue
    i=i+1
    if(SortCharacter(indx(i)) < a)goto 3
4   continue
    j=j-1
    if(SortCharacter(indx(j)) > a)goto 4
    if(j < i)goto 5
    itemp=indx(i)
    indx(i)=indx(j)
    indx(j)=itemp
    goto 3
5   indx(l+1)=indx(j)
    indx(j)=indxt
    jstack=jstack+2
    if(jstack > NSTACK) THEN
      WRITE(*, *) 'NSTACK too small in indexx'
      WRITE(*,*)'Program paused. Press ENTER to continue.'
      READ(*,*)
    END IF
    if(ir-i+1 >= j-l)then
      istack(jstack)=ir
      istack(jstack-1)=i
      ir=j-1
    else
      istack(jstack)=j-1
      istack(jstack-1)=l
      l=i
    endif
  endif
  goto 1
END SUBROUTINE indexCh  
!
FUNCTION gammln(xx)
  DOUBLE PRECISION gammln,xx
  INTEGER j
  DOUBLE PRECISION ser,stp,tmp,x,y,cof(6)
  SAVE cof,stp
  DATA cof,stp/76.18009172947146d0,-86.50532032941677d0, &
   24.01409824083091d0,-1.231739572450155d0,.1208650973866179d-2, &
   -.5395239384953d-5,2.5066282746310005d0/
  x=xx
  y=x
  tmp=x+5.5d0
  tmp=(x+0.5d0)*log(tmp)-tmp
  ser=1.000000000190015d0
  do j=1,6
    y=y+1.d0
    ser=ser+cof(j)/y
  end do
  gammln=tmp+log(stp*ser/x)
  return
END FUNCTION gammln
!
SUBROUTINE StopOnError(string)
  CHARACTER(LEN=*), INTENT(IN) :: string
  write (*,*) 'StopOnError: ',string
  STOP 'program terminated by StopOnError'
END SUBROUTINE StopOnError